from haplotyping._version import __version__
__author__ = "Matthijs Brouwer"

from haplotyping.data.validator import *
from haplotyping.data.constructDatabase import *
from haplotyping.data.checkDatabase import *